#include "group_definitions.h"
#include "ui_group_definitions.h"
#include <stdio.h>
#include "qmolcalc.h"
#include <QFile>
#include <QTextStream>
#include <QDebug>
#include <QMessageBox>
#include <QStandardPaths>
#include <QStatusBar>

group_definitions::group_definitions(QWidget *parent, ElementTable* elements_, GroupTable* groups_, qmolPreferences* preferences_):
    QDialog(parent),
    ui(new Ui::group_definitions)
{
    preferences = preferences_;
    QList<int> splitterSizes;
    splitterSizes << 90 << 250;
    ui->setupUi(this);
    resize(preferences->groupDefsSize);
    move(preferences->groupDefsPosition);
    ui->splitter->setSizes(splitterSizes);
    ui->groupTableWidget->setRowCount(0);
    ui->groupTableWidget->setColumnCount(2);
    ui->groupTableWidget->horizontalHeader()->setSectionResizeMode(QHeaderView::ResizeToContents);
    ui->groupTableWidget->setHorizontalHeaderItem(0, new QTableWidgetItem(tr("Symbol")));
    ui->groupTableWidget->setHorizontalHeaderItem(1, new QTableWidgetItem(tr("Sum")));
    ui->pbDeleteLine->setEnabled(false);

    ui->cbAlkylsAryls->setChecked(preferences->activePredefSets.value(m_AlkylsAryls));
    ui->cbAminoAcids->setChecked(preferences->activePredefSets.value(m_AminoAcids));
    //ui->cbAminoAcids1L->setChecked(preferences->activePredefSets.value(m_AminoAcids1L));

    elements = elements_;
    groups = groups_;
    connect(this, SIGNAL(tableChanged(bool)), this, SLOT(onTableChanged(bool)));
    connect(ui->groupTableWidget, SIGNAL(itemChanged(QTableWidgetItem *)), this, SLOT(onItemChanged(QTableWidgetItem *)));
    connect(ui->groupTableWidget, SIGNAL(itemSelectionChanged()), this, SLOT(onSelectionChanged()));
    composition = new Composition;
    changed = false;
}

group_definitions::~group_definitions()
{
    delete ui;
    delete composition;
}

void group_definitions::on_pbDeleteLine_clicked() {
    int rows = linesToDelete.count();
    for(int i = rows - 1; i >= 0; i--){
        int currentRow = linesToDelete.value(i);
        QString symbolToRemove = ui->groupTableWidget->item(currentRow, 0)->text();
        ui->groupTableWidget->removeRow(currentRow);
        if (groups->isOverriding(symbolToRemove)) {
            elements->undoOverride(symbolToRemove);
        }
    }
    QString selectionStatus = tr("Deleted %n lines.", "", rows);
    changed = true;
    emit tableChanged(changed);
    ui->lbStatus->setText(selectionStatus);
}

void group_definitions::onShow() {
    ui->symbol->setFocus();
}

void group_definitions::onTableChanged(bool changed_) {
    ui->pbApply->setEnabled(changed_);
    ui->pbDiscard->setEnabled(changed_);
}

void group_definitions::onSelectionChanged() {
    QTableWidgetSelectionRange selectionRange;
    QList<QTableWidgetSelectionRange> selectedRanges = ui->groupTableWidget->selectedRanges();
    QString selectionStatus = "";

    int toDelete = 0;
    int selectedRows, selectedCols, topRow;

    linesToDelete.clear();

    if (selectedRanges.isEmpty()) {
        ui->pbDeleteLine->setEnabled(false);
        ui->lbStatus->setText("");
        return;
    }

    foreach (selectionRange, selectedRanges) {
        topRow = selectionRange.topRow();
        selectedRows = selectionRange.rowCount();
        toDelete += selectedRows;

        for (int i = 0; i < selectedRows; i++) {
            linesToDelete.append(topRow + i);
        }
    }

    if (toDelete!=0) {
        ui->pbDeleteLine->setEnabled(true);
        selectionStatus = tr("Lines selected for deletion: %n", "", toDelete);
    }

    ui->lbStatus->setText(selectionStatus);
    qSort(linesToDelete);
}

void group_definitions::predefToggled() {
    changed = true;
    emit tableChanged(changed);
}

void group_definitions::on_cbAlkylsAryls_toggled(bool checked) { predefToggled(); }
void group_definitions::on_cbAminoAcids_toggled(bool checked) { predefToggled(); }
//void group_definitions::on_cbAminoAcids1L_toggled(bool checked) { predefToggled(); }

void group_definitions::onItemChanged(QTableWidgetItem *item) {
    changed = true;
    int row = item->row() + 1;
    QString errorStatus = "";

    int* errors = new int[3];
    if (item->column() == 1) {
        QString inputFormula = item->text();
        if (inputFormula == NULL) return;

        if (checkItem(item, errors))
            ui->lbStatus->setText(getStatusMessage(errors));
        else
            ui->lbStatus->setText("");
    }
    changed = true;
    emit tableChanged(changed);
    //ui->lbStatus->setText("Table changed.");
    delete [] errors;
}

void group_definitions::on_pbApply_clicked(){
    QString errorStatus = writeTable();

    int* errors = new int[3];
    if (checkTable(errors)) {
        ui->lbStatus->setText(getStatusMessage(errors));
    } else {
        ui->lbStatus->setText(tr("Changes applied."));
    }

    preferences->activePredefSets.replace(m_AlkylsAryls,  ui->cbAlkylsAryls->isChecked());
    preferences->activePredefSets.replace(m_AminoAcids,   ui->cbAminoAcids->isChecked());
    //preferences->activePredefSets.replace(m_AminoAcids1L, ui->cbAminoAcids1L->isChecked());

    changed = false;
    emit tableChanged(changed);

    groups->readTables();
    delete [] errors;
}

void group_definitions::on_pbDiscard_clicked(){
    groups->readTables();
    deleteTableContents();
    setTableContents();
    changed = false;
    emit tableChanged(changed);
}

void group_definitions::on_pbCancel_clicked(){
    on_pbDiscard_clicked();
    hide();
}

void group_definitions::on_pbDone_clicked(){
    if (changed) on_pbApply_clicked();
    ui->symbol->setFocus();
    hide();
}

void group_definitions::addNewDefinition() {
    addGroup();
    changed = true;
}

void group_definitions::on_definition_returnPressed() {
    addNewDefinition();
}

void group_definitions::on_symbol_returnPressed() {
    if (eatWhiteSpace(ui->symbol->text())!="") {
        if (eatWhiteSpace(ui->definition->text())=="") {
            ui->definition->setFocus();
        } else {
            addNewDefinition();
        }
    }
}

void group_definitions::addGroup() {
    QString symbol, definition, newDefinition, compString, coeff, line;
    QList<QTableWidgetItem *> found;
    int row, rows = ui->groupTableWidget->rowCount();
    symbol = eatWhiteSpace(ui->symbol->text());
    definition = eatWhiteSpace(ui->definition->text());
    if (symbol=="") {
        ui->lbStatus->setText(tr("Enter a symbol name!"));
        return;
    }
    if (definition=="") {
        ui->lbStatus->setText(tr("Enter a group definition!"));
        return;
    }

    found = ui->groupTableWidget->findItems(symbol, Qt::MatchExactly);
    if (!found.isEmpty()) {
        ui->lbStatus->setText(tr("Group %1 is aleady defined.").arg(symbol));
        return;
    }

    //if (elements->contains(symbol)) {
    //     ui->lbStatus->setText(tr("%1 is an element symbol.").arg(symbol));
    //     return;
    //}

    qMolCalc* qmolcalc = new qMolCalc(elements, groups);
    newDefinition = qmolcalc->parseFormula(definition);
    delete qmolcalc;


    QTableWidgetItem* definitionItem = new QTableWidgetItem(definition);

    int* errors = new int[3];
    if (checkItem(definitionItem, errors)) {
        errors[m_Row] = m_toAdd;
        ui->lbStatus->setText(getStatusMessage(errors));
        delete definitionItem;
        return;
    };

    ui->groupTableWidget->insertRow(rows);
    ui->groupTableWidget->setItem(rows, 0, new QTableWidgetItem(symbol));
    ui->groupTableWidget->setItem(rows, 1, definitionItem);//eatWhiteSpace(newDefinition)
    ui->symbol->clear();
    ui->definition->clear();
    ui->symbol->setFocus();

    delete [] errors;
}

bool group_definitions::checkTable(int* errors) {
    int row, rows = ui->groupTableWidget->rowCount();
    QTableWidgetItem* item;

    for (row = 0 ; row < rows ; row++) {
        item = ui->groupTableWidget->item(row, 1);
        if (checkItem(item, errors)) return true;
    }
    return false;
}

bool group_definitions::checkItem(QTableWidgetItem* item, int* errors) {
    errors[m_Error] = m_OK;
    QString definition, sumFormula;
    double groupWeight;
    bool isNumber;
    definition = item->text();
    errors[m_Row] = item->row();

    groupWeight = definition.toDouble(&isNumber);

    if (isNumber) {
        errors[m_Kind] = m_Weight;
    } else {
        errors[m_Kind] = m_Formula;
        qMolCalc* qmolcalc = new qMolCalc(elements, groups);
        sumFormula = qmolcalc->parseFormula(definition);
        delete qmolcalc;

        if (sumFormula.contains(tr("Undefined"))) {
            errors[m_Error] = m_UndefinedSymbol;
        } else
        if (sumFormula.contains(tr("Bad formula"))) {
            errors[m_Error] = m_BadFormula;
        }
    }

    if (errors[m_Error] == m_OK)
        return false;
    else
        return true;
}

QString group_definitions::getStatusMessage(int* errors){
    if (errors[m_Row] == m_toAdd) {
        switch (errors[m_Error]) {
            case m_UndefinedSymbol:
                return tr("Recheck your definition for unknown symbols.");
                break;
            case m_BadFormula:
                return tr("Bad formula.");
                break;
        }
    } else {
        int row = errors[m_Row] + 1;
        switch (errors[m_Error]) {
            case m_UndefinedSymbol:
                return tr("Warning: row %n has unknown symbols in definition.", "", row);
                break;
            case m_BadFormula:
                return tr("Warning: row %n contains a bad formula.", "", row);
                break;
        }
    };
    return "";
}

QString group_definitions::writeTable() {
    QString symbol, definition, sumFormula, coeff, line;
    QString errorStatus = "ok";
    QString dataPath = QStandardPaths::standardLocations(QStandardPaths::DataLocation).join("");
    int row, rows = ui->groupTableWidget->rowCount();

    QFile grpFile(dataPath + "/groups");

    grpFile.open(QIODevice::WriteOnly);

    QTextStream grpFileStream(&grpFile);

    for (row = 0 ; row < rows ; row++) {
       symbol = ui->groupTableWidget->item(row, 0)->text();
       definition = ui->groupTableWidget->item(row, 1)->text();

       grpFileStream << symbol << " " << definition << "\n";
    }

    grpFile.close();
    return errorStatus;
}

void group_definitions::deleteTableContents(){
    int  rows = ui->groupTableWidget->rowCount();
    for(int i = rows - 1; i >= 0; i--){
        ui->groupTableWidget->removeRow(i);
    }
}

void group_definitions::setTableContents(){
    QHashIterator<QString, QString> groupsFormulaIt(*(groups->getGrpFormulaTable()));
    QHashIterator<QString, double> groupsWeightIt(*(groups->getGrpWeightTable()));
    int row = 0;
    deleteTableContents();

    while (groupsFormulaIt.hasNext()){
         groupsFormulaIt.next();
         ui->groupTableWidget->insertRow(row);
         ui->groupTableWidget->setItem(row, 0, new QTableWidgetItem(groupsFormulaIt.key()));
         ui->groupTableWidget->setItem(row++, 1, new QTableWidgetItem(eatWhiteSpace(groupsFormulaIt.value())));
    }
    while (groupsWeightIt.hasNext()){
         groupsWeightIt.next();
         ui->groupTableWidget->insertRow(row);
         ui->groupTableWidget->setItem(row, 0, new QTableWidgetItem(groupsWeightIt.key()));
         ui->groupTableWidget->setItem(row++, 1, new QTableWidgetItem(QString::number(groupsWeightIt.value())));
    }

    int* errors = new int[3];
    if (checkTable(errors)) {
        ui->lbStatus->setText(getStatusMessage(errors));
    } else {
        ui->lbStatus->setText(tr("%n group definitions loaded.", "", row));
    }

    ui->lbStatus->show();
    changed = false;
    emit tableChanged(false);
    delete [] errors;
}
