/*
 * qmol.cpp 
 *
 * This file is part of qmol - A simple molecular weight calculator
 *
 * Derivative of Tomislav Gountchev's original program kmol
 * (Idea, basic layout, formula parser)
 *
 * Rewrite based on Qt4 (i.e. no KDE dependencies) 
 * + Added a history feature
 * + polished the group definition dialog
 *
 * Copyright (C) 2012 Thomas Mitterfellner
 * Copyright (C) 2009 Tomislav Gountchev
 *
 */

/***************************************************************************
 *                                                                         *
 *   qmol is free software; you can redistribute it and/or modify          *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful, but   *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the Free Software           *
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,            *
 *   MA  02110-1301, USA                                                   *
 *                                                                         *
 *                                                                         *
 ***************************************************************************/

#include <QMessageBox>
#include <QSettings>
#include <QDesktopServices>
#include <QUrl>
#include <QFile>
#include <QStyledItemDelegate>
#include <QAbstractTextDocumentLayout>
#include <QDebug>

#ifdef Q_OS_UNIX
  #include "cmake_defines.h" // this is generated by cmake and defines the handbook installation path
#endif
#include "qmol.h"
#include "ui_qmol.h"
#include "qmolcalc.h"
#include "stdio.h"
#include "options_dialog.h"
#include "preferences.h"

qmol::qmol(QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::qmol)
{
    ui->setupUi(this);
    ui->composition->setReadOnly(true);

    preferences = new qmolPreferences;
    preferences->restoreSettings();
    preferences->restoreGUISettings();
    resize(preferences->size);
    move(preferences->position);

    overrides = new QSet<QString>;
    elements = new ElementTable;
    groups = new GroupTable(elements, preferences);
    groupDefsWindow = new group_definitions(0, elements, groups, preferences);
    aboutWindow = new aboutDialog();
    qmolcalc = new qMolCalc(elements, groups, preferences);
    model = new QStandardItemModel(0,4);
    ui->historyTableView->setModel(model);
    delegate = new tableDelegate;
    ui->historyTableView->setItemDelegate(delegate);
    delegate->historyTableView = ui->historyTableView;
    completer = new QCompleter(model);

    model->setHeaderData( 0, Qt::Horizontal, QObject::tr("Formula"));
    model->setHeaderData( 1, Qt::Horizontal, QObject::tr("Sum"));
    model->setHeaderData( 2, Qt::Horizontal, QObject::tr("Weight"));


    ui->historyTableView->setEditTriggers(QAbstractItemView::NoEditTriggers);
    ui->historyTableView->setColumnHidden(3, true);
    ui->pbRemoveSelected->setEnabled(false);

    QFontMetrics fm(font());
    int h = qRound(qMax(fm.height(), 14) * 1.2) ;
    ui->historyTableView->verticalHeader()->setDefaultSectionSize(h);

    //connect(ui->historyTableView, SIGNAL(clicked(QModelIndex)), this, SLOT(onSelectionChanged()));
    connect(ui->historyTableView->selectionModel(), SIGNAL(selectionChanged(const QItemSelection, const QItemSelection)), this, SLOT(onSelectionChanged(const QItemSelection, const QItemSelection)));

    ui->cbReverseHistory->setChecked(preferences->reverseHistory);
    ui->splitter->restoreState(preferences->splitterSizes);
    //ui->splitter->hide();
    ui->formula->setCompleter(completer);
    /* QFont font_ = font();
    font_.setPointSize(14);
    setFont(font_); */
}

qmol::~qmol()
{
    delete ui;
    delete preferences;
    delete groupDefsWindow;
    delete aboutWindow;
    delete overrides;
    delete elements;
    delete groups;
    delete model;
    delete delegate;
    delete completer;
    delete qmolcalc;
}


void qmol::quit(){
    preferences->size = size();
    preferences->position = pos();
    preferences->splitterSizes = ui->splitter->saveState();
    preferences->groupDefsPosition = groupDefsWindow->pos();
    preferences->groupDefsSize = groupDefsWindow->size();

    preferences->storeSettings();

    QApplication::quit();
}

void qmol::showAbout(){
    aboutWindow->exec();
}

void qmol::openHandbook(){
    QString handbookPath;
    QUrl handbookUrl;
    QFile file;

    #ifdef Q_OS_LINUX
        handbookPath = HANDBOOK_PATH"qmol_handbook.html";
        handbookUrl = QUrl::fromLocalFile(handbookPath);
    #else
        handbookPath = QCoreApplication::applicationDirPath() + "/docs/qmol_handbook.html";
        handbookUrl = QUrl::fromLocalFile(handbookPath);
    #endif

    file.setFileName(handbookPath);

    if (!(file.exists())) {
        handbookPath = "qmol.sourceforge.net/docs/qmol_handbook.html";
        handbookUrl = QUrl::fromUserInput(handbookPath);
    }

    QDesktopServices::openUrl(handbookUrl);
}

void qmol::onSelectionChanged(const QItemSelection& newSelection, const QItemSelection& oldSelection){

    QItemSelectionModel *select = ui->historyTableView->selectionModel();
    QList<QModelIndex> selectionList = select->selectedRows(1);
    QList<QModelIndex> selectionList0 = select->selectedRows(0);
    QListIterator<QModelIndex> selectionIt(selectionList);
    QListIterator<QModelIndex> selectionIt0(selectionList0);

    QString formula = "";
    QString formula0 = "";
    QString separator = "";
    QString weight;
    QModelIndex index, index0;
    //Composition comp;

    if (selectionList.isEmpty()) {
        ui->pbRemoveSelected->setEnabled(false);
    } else {
        ui->pbRemoveSelected->setEnabled(true);

        qmolcalc->clearComp();

        while (selectionIt.hasNext()) {
            index = selectionIt.next();
            index0 = selectionIt0.next();
            if (selectionIt.hasNext())
                  separator = "+";
            else
                  separator = "";
            formula += index.data(Qt::DisplayRole).toString() + separator;
            formula0 += index0.data(Qt::DisplayRole).toString() + separator;
        }

        qmolcalc->parseFormula(formula0);

        ui->formula->setText(formula0);
        weight.setNum(qmolcalc->getFormulaWeight());
        ui->result->setText(weight);

        qmolcalc->printFullComposition(ui->composition, preferences->outputFormatGUI);
    }
}

void qmol::calc(){
    QString inputFormula = ui->formula->text();
    QString error = qmolcalc->parseFormula(inputFormula);
    QStringList fullComposition;
    //QString fullComposition;

    if (error == c_OK) {
        double weight = qmolcalc->getFormulaWeight();
        QString result; result.setNum(weight);
        ui->result->setText(result);
        qmolcalc->printFullComposition(ui->composition, preferences->outputFormatGUI);
        addToHistory(inputFormula);
    } else {
        ui->result->setText(error);
    }
}

tableDelegate::tableDelegate(QObject *parent)
     : QStyledItemDelegate(parent) {
}

void tableDelegate::paint(QPainter *painter, const QStyleOptionViewItem &option, const QModelIndex &index) const
{
     QTextDocument textDoc;
     QVariant text = index.data(Qt::DisplayRole);
     QString textHTML;
     QStyleOptionViewItemV4 optionV4 = option;
     QStyle *style = QApplication::style();
     QRect clipRect = style->subElementRect(QStyle::SE_ItemViewItemText, &optionV4);
     QPoint point = option.rect.topLeft();

     initStyleOption(&optionV4, index);

     //QStyle *style = optionV4.widget ? optionV4.widget->style() : QApplication::style();

     // Painting item without text
     optionV4.text = QString();
     style->drawControl(QStyle::CE_ItemViewItem, &optionV4, painter);

     //point.setY(point.y());
     //point.setX(point.x()+2);
     //painter->setClipRect(clipRect);

     textHTML = text.toString();

     // for column 1 (sum formula), get data from hidden column 3
     if (index.column()==1)
         if (!textHTML.isEmpty())
             textHTML = historyTableView->model()->index(index.row(), index.column() + 2).data(Qt::DisplayRole).toString();

     textDoc.setHtml(textHTML);

     QAbstractTextDocumentLayout::PaintContext ctx;
     int rowHeight = historyTableView->rowHeight(index.row());
     int contentHeight = textDoc.size().height();
     int dY = int(rowHeight - contentHeight)/2;
     point.setY(point.y()+dY);

     painter->save();
     painter->translate(point);
     painter->setClipRect(clipRect.translated(-clipRect.topLeft()));
     textDoc.documentLayout()->draw(painter, ctx);
     painter->restore();
}

void qmol::reverseHistory() {
    int rows = model->rowCount();
    const int halfOfRows = (int)rows / 2;

    preferences->reverseHistory = ui->cbReverseHistory->isChecked();

    for (int i = 0 ; i < halfOfRows ; i++) {

        int sourceRow = i;
        int destRow = rows - i - 1;
        if (sourceRow == (destRow+1)) break;

        // take the rows from the table
        QString sourceHeaderString = model->headerData(sourceRow, Qt::Vertical).toString();
        QList<QStandardItem*> sourceItems = model->takeRow(sourceRow);
        //QStandardItem* sourceHeader = model->takeVerticalHeaderItem(sourceRow);
        // now there's one line less, therefore take destRow - 1
        QString destHeaderString = model->headerData(destRow - 1, Qt::Vertical).toString();
        QList<QStandardItem*> destItems = model->takeRow(destRow - 1);
        //QStandardItem* destHeader = model->takeVerticalHeaderItem(destRow - 1);

        // set back in reverse order
        model->insertRow(sourceRow, destItems);
        model->insertRow(destRow, sourceItems);

        //model->setHeaderData(sourceRow, Qt::Vertical, QString::number(destRow + 1));
        //model->setHeaderData(destRow, Qt::Vertical, QString::number(sourceRow + 1));
        model->setHeaderData(sourceRow, Qt::Vertical, destHeaderString);
        model->setHeaderData(destRow, Qt::Vertical, sourceHeaderString);
    }
    //ui->historyTableView->repaint();
}

void qmol::addToHistory(QString formula) {
    QString weight, sumFormula, sumFormulaHTML;
    bool reverseHistory = ui->cbReverseHistory->isChecked();
    QList<QStandardItem *> foundItems = model->findItems(formula, Qt::MatchExactly, 0);

    if (foundItems.isEmpty()) {
        int rows = model->rowCount();
        //weight = QString::number(qmolcalc->getFormulaWeight(), 'f', 4);
        weight.setNum(qmolcalc->getFormulaWeight());
        sumFormula = eatWhiteSpace(qmolcalc->getSumFormula());
        sumFormulaHTML = qmolcalc->getSumFormulaHTML();

        QList<QStandardItem*> newRow;
        /* column 0: formula as entered; 1: sum formula (plain text); 2: formula weight;
         * 3: hidden column containing the HTML representation of col 2 */
        newRow << new QStandardItem(formula) << new QStandardItem(sumFormula) << new QStandardItem(weight)
               << new QStandardItem(sumFormulaHTML);

        if (reverseHistory) {
            model->insertRow(0, newRow);
            model->setHeaderData( 0, Qt::Vertical, QString::number(rows + 1));
        } else {
            model->appendRow(newRow);
            model->setHeaderData( rows, Qt::Vertical, QString::number(rows + 1));
        }

        ui->historyTableView->resizeColumnsToContents();
    }
}

void qmol::refreshHistory(){
    QString weight, sumFormula, sumFormulaHTML;
    int rowCount = model->rowCount();

    for (int i = 0 ; i < rowCount ; i++) {
        QString headerString = model->headerData(i, Qt::Vertical).toString();
        QString formula = model->item(i, 1)->data(Qt::DisplayRole).toString();
        QString formulaHTML = qMolCalc::plainFormulaToHTML(formula, preferences->coeffColor, preferences->subscriptCoeffient);
        QStandardItem* formulaHTMLitem = model->item(i, 3);
        formulaHTMLitem->setData(formulaHTML, Qt::DisplayRole);
        model->setItem(i, 3, formulaHTMLitem);
    }
}

void qmol::openDefineDialog(){
    groupDefsWindow->setTableContents();
    groupDefsWindow->exec();
}

void qmol::openOptionsDialog(){
    options_dialog optionsDialog(preferences);
    optionsDialog.exec();

    preferences->preferencesSize = optionsDialog.size();
    preferences->preferencesPosition = optionsDialog.pos();

    qmolcalc->setOutputFormat(preferences->outputFormat);
    refreshHistory();
}

void qmol::on_clear_fields_clicked()
{
    ui->formula->clear();
    ui->result->clear();
    ui->composition->clear();
}

void qmol::on_calculate_clicked()
{
    ui->result->clear();
    ui->composition->clear();
    calc();
}

void qmol::clearHistory() {
    int rows = ui->historyTableView->model()->rowCount();
    ui->historyTableView->model()->removeRows(0, rows);
}

void qmol::removeFromHistory() {
    QItemSelectionModel *select = ui->historyTableView->selectionModel();
    QList<QModelIndex> selectionList = select->selectedRows(1);
    bool reverseHistory = ui->cbReverseHistory->isChecked();
    int selectedItems = selectionList.count();
    int rows, j;
    QList<int> selectedRows;
    //QList<QString> selectionList = selected.indexes();
    QListIterator<QModelIndex> selectionIt(selectionList);

    QModelIndex index;

    while (selectionIt.hasNext()) {
        index = selectionIt.next();
        selectedRows.append(index.row());
    }

    qSort(selectedRows);

    for (int i = selectedItems - 1 ; i >= 0 ; i-- ) {
        int currentRow = selectedRows.at(i);
        QString itemString = ui->historyTableView->model()->index(currentRow, 0).data(Qt::DisplayRole).toString();
        ui->historyTableView->model()->removeRow(currentRow);
        //historyContent->removeOne(itemString);
    }
    rows = ui->historyTableView->model()->rowCount();
    for (int i = 0 ; i < rows ; i++) {
        if (reverseHistory)
            j = rows - i;
        else
            j = i + 1;
        model->setHeaderData(i, Qt::Vertical, j);
    }
}

void qmol::on_formula_returnPressed() { on_calculate_clicked(); }
