/*
 * qmolcalc.cpp
 *
 * This file is part of qmol - A simple molecular weight calculator
 *
 * Copyright (C) 2012 Thomas Mitterfellner
 * Copyright (C) 2009 Tomislav Gountchev
 *
 */

/***************************************************************************
 *                                                                         *
 *   qmol is free software; you can redistribute it and/or modify          *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful, but   *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the Free Software           *
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,                 *
 *   MA 02111-1307 USA                                                     *
 *                                                                         *
 ***************************************************************************/

#include <QTextStream>
#include <QMessageBox>
#include <QDebug>

#include "qmolcalc.h"

qMolCalc::qMolCalc(ElementTable* elements_, GroupTable* groups_, qmolPreferences* preferences_) {

    proteinMode = false;

    if (preferences_ != NULL) {
        preferences = preferences_;
        outputFormat = preferences->outputFormat;
        proteinMode = preferences->proteinMode;
    } else preferences = NULL;

    deleteElementsTable = deleteGroupsTable = false;
    if (elements_ == NULL) {
        elements = new ElementTable;
        deleteElementsTable = true;
    } else { elements = elements_; }
    if (groups_ == NULL) {
        groups = new GroupTable(elements, preferences);
        deleteGroupsTable = true;
    } else  groups = groups_;

    composition = new Composition();
    percent = new QList<double>;
    fullComposition = new QStringList();
}

qMolCalc::~qMolCalc() {
    delete composition;
    delete percent;
    delete fullComposition;
    if (deleteElementsTable) delete elements;
    if (deleteGroupsTable) delete groups;
}

void qMolCalc::setOutputFormat(QString outputFormat_){
   outputFormat = outputFormat_;
}

double qMolCalc::getWeight() const {
    /* Wenn in elements nicht gefunden, dann in groups schauen */
    //return elements->getWeight(name);
    //return calcFormulaWeight(name);
    return 0;
}

double qMolCalc::getFormulaWeight() const {
    return formulaWeight;
}

QString eatWhiteSpace(QString string) {
    QTextStream inStream(&string);
    QString outString, nonWhiteSpace;
    while (true) {
        inStream >> ws >> nonWhiteSpace;
        if (nonWhiteSpace=="") break;
        outString.append(nonWhiteSpace);
    }
    return outString;
}

void qMolCalc::makeFullComposition(QString outputFormat_) {
    QMapIterator<QString, double> listIt(*(composition->getList()));
    QString symbol, line, weight;
    int numberAminoAcids;
    QString c_numberAminoAcids = "";
    const QString compositionFormat = preferences->compositionFormat;
    const QString numberAAformat = preferences->numberAAformat;
    double percentValue;
    int i = 0;

    outputFormat = outputFormat_;

    calcPercentages();
    fullComposition->clear();
    weight.setNum(formulaWeight);

    commandLineOutput = outputFormat; //preferences->outputFormat;

    compositionOneLine.clear();

    while (listIt.hasNext()) {
         listIt.next();
         percentValue = percent->value(i++);
         symbol = listIt.key();
         symbol.remove(c_overriddensymbol[0]);
         line = compositionFormat;
         line.replace("%E", symbol);
         line.replace("%C", QString::number(percentValue,'f',2));
         fullComposition->append(line);
         compositionOneLine.append(line);
    }
    // remove last \n if present to avoid empty line not explicitly stated
    if (compositionOneLine.endsWith("\\n")) compositionOneLine.chop(2);

    // prepare number (sum) of amino acids output
    numberAminoAcids = composition->getNumberAminoAcids();
    if (numberAminoAcids != 0) {
        c_numberAminoAcids = numberAAformat;
        c_numberAminoAcids.replace("%N", QString::number(numberAminoAcids));
    }


    // prepare full output
    // substitutions common to GUI and command line output:
    commandLineOutput.replace("%i", inputFormula);
    commandLineOutput.replace("%w", weight);
    if (hasNegativeCoefficients())
        commandLineOutput.replace("%c", QString(QObject::tr("Formula has negative coefficients")));
    else
        commandLineOutput.replace("%c", compositionOneLine);
    commandLineOutput.replace("%a", c_numberAminoAcids);

    // tabs and spaces are replaced after common block:
    commandLineOutput.replace("\\n", "\n");
    commandLineOutput.replace("\\t", "\t");

    guiOutput = commandLineOutput;


    // special substitutions for GUI/command line:
    commandLineOutput.replace("%o", eatWhiteSpace(sumFormula));
    guiOutput.replace("%o", sumFormulaHTML);

}

void qMolCalc::printFullComposition(QTextEdit* compositionWidget, QString outputFormat_, int outputMode) {
    QString line;
    QStringList fullComposition;

    makeFullComposition(outputFormat_);

    fullComposition = getResultsOutput(outputMode).split("\n");

    compositionWidget->clear();
    foreach (line, fullComposition) compositionWidget->append(line);
}


Composition* qMolCalc::getComposition() { return composition; }

QStringList* qMolCalc::getFullComposition() {
    makeFullComposition("%c");
    return fullComposition;
}

bool qMolCalc::hasNegativeCoefficients() {
    QMap<QString, double>* composition_ = composition->getList();
    foreach (double coefficient, composition_->values()) {
        if (coefficient < 0) return true;
    }
    return false;
}

QString qMolCalc::getResultsOutput(int outputMode, bool createResults) {
    QString outputFormat_;
    switch (outputMode) {
        case m_GUI:
            return guiOutput;
            break;
        case m_CommandLine:
            if (createResults) makeFullComposition(preferences->outputFormat);
            return commandLineOutput;
            break;
    }
}

void qMolCalc::calcPercentages() {
    QMapIterator<QString, double> listIt(*(composition->getList()));
    double coeff, elementWeight;

    percent->clear();

    while (listIt.hasNext()) {
        listIt.next();
        elementWeight = elements->getWeight(listIt.key());
        coeff = listIt.value();
        percent->append(100 * coeff * elementWeight / formulaWeight);
    }
}

void qMolCalc::calcWeight() {
    QMapIterator<QString, double> listIt(*(composition->getList()));
    double coeff, elWeight;
    formulaWeight = 0;

    while(listIt.hasNext()) {
        listIt.next();
        elWeight = elements->getWeight(listIt.key());
        coeff = listIt.value();
        formulaWeight +=  elWeight * coeff;
    }
}

void qMolCalc::makeSumFormula() {
    QMapIterator<QString, double> listIt(*(composition->getList()));
    QString coeff;

    sumFormula = "";

    while (listIt.hasNext()) {
        listIt.next();
        sumFormula += listIt.key();
        sumFormula += " ";

        double value = listIt.value();
        coeff.setNum(value);

        if (coeff != "1") sumFormula += coeff + " ";
    }

    makeSumFormulaHTML(sumFormula);

}

QString qMolCalc::plainFormulaToHTML(QString s, QColor color, bool subscriptCoeff) {
    int i = 0, j = 0;
    QString symbol, coeff = "", coeffHTML = "";
    QString formulaHTML = "";
    QString coeffColor = color.name();
    int sl = s.length();
    QString coeffOpenTag, fontOpenTag, subOpenTag;
    QString coeffCloseTag, fontCloseTag, subCloseTag;

    fontOpenTag = "<font color=\"" + coeffColor + "\">";
    fontCloseTag = "</font>";
    subOpenTag = subscriptCoeff ? "<sub>" : "";
    subCloseTag = subscriptCoeff ? "</sub>" : "";

    coeffOpenTag = subOpenTag + fontOpenTag;
    coeffCloseTag = fontCloseTag + subCloseTag;

    while (j < sl) {

        while (i < sl && ((s[i] >= 'A' && s[i] <= 'Z')
                      || (s[i] >= 'a' && s[i] <= 'z')
                      || (s[i] == '*' || s[i] == '\''
                      ||  s[i] == c_overriddensymbol[0]))) i++;

        symbol = s.mid(j, i-j);
        symbol.remove(c_overriddensymbol[0]);

        j = i;

        while (i < sl && ((s[i] <= '9' && s[i] >= '0') || s[i] == '.' || s[i] == '-')) i++;
        coeff = s.mid(j, i-j);
        j = i;
        if (!coeff.isEmpty())
            coeffHTML = coeffOpenTag + coeff + coeffCloseTag;
        else
            coeffHTML = "";

        formulaHTML += symbol + coeffHTML;
    }

    return formulaHTML;
}



void qMolCalc::makeSumFormulaHTML(QString s) {
    if (preferences != NULL) {
        QColor coeffColor = preferences->coeffColor;
        bool subscriptCoeff = preferences->subscriptCoeffient;
        sumFormulaHTML = plainFormulaToHTML(eatWhiteSpace(s), coeffColor, subscriptCoeff);
    }
}

QString qMolCalc::getSumFormula() const {
    //makeSumFormula();
    return sumFormula;
}

QString qMolCalc::getSumFormulaHTML() const {
    //makeSumFormula();
    return sumFormulaHTML;
}

double qMolCalc::calculateFormulaWeight(QString formula, ElementTable* elementTable) {
    QString name;
    double coef, weight = 0, elWeight, grpWeight, tokenWeight;
    //ElementTable elTable;
    //GroupTable grpTbl;
    QTextStream in(&formula);

    while (true) {
        in >> name >> ws >> coef >> ws;
        if (name=="") break;
        if (coef==0) coef=1;
        elWeight = elementTable->getWeight(name);
        weight += coef * elWeight;
    }
    return weight;
}

void qMolCalc::clearComp() {
    composition->list->clear();
}

bool qMolCalc::compIsEmpty() const {
    return composition->list->isEmpty();
}

QString qMolCalc::parseFormula(const QString& s, Composition* comp) {
    QString formula = eatWhiteSpace(s);
    if (comp == NULL) {
        comp = composition;
        comp->clear();
    }

    proteinMode = preferences != NULL ? preferences->proteinMode : false;
    QString error = parseString(formula, comp);
    inputFormula = formula;

    if (error=="OK") {
        calcWeight();
        makeSumFormula();
    }
    return error;
}

QString qMolCalc::checkBrackets(const QString& s, int sl, int* i, int j, QChar bracket) {
    int level = 1;
    QChar closingBracket = (bracket == '[') ? ']' : ')';
    while (true) {
        if ((*i)++ == sl) {
            return QString(QObject::tr("Bad formula.")); // ERROR
        }
        if (s[*i] == bracket) level++;
        if (s[*i] == closingBracket) level--;
        if (level == 0) break;
    }
    if (*i == (j+1)) { //
        return QString (QObject::tr("Empty brackets.")); // ERROR
    }
    return c_OK;
}

QString qMolCalc::parseString(const QString& s, Composition* comp, bool aaMode, int brace) {
    if (s.isEmpty()) return QString (QObject::tr("Enter a formula.")); //ERROR
    int sl = s.length();
    int i = 0;
    int k = 0;
    QString errors (c_OK);
    bool ok = true;
    int sign = 1;

    switch (s[i].toLatin1()) {
        case '-': sign = -1;
        case '+': i++; k++;
    }

    // Parse for prefix
    while (i < sl && ((s[i] <= '9' && s[i] >= '0') || s[i] == '.')) i++;
    double prefix = sign * (i == k ? 1 : s.mid(k,i-k).toDouble(&ok));
    if (! ok || i == sl) return QString (QObject::tr("Bad formula.")); // ERROR*/
    //if (prefix == 0) return errors; // no matter what will follow, if prefix==0 it won't contribute to the weight

    Composition compTemp;
    while (i < sl) {
        int j=i;

        if (s[i] == '[' || s[i] == '(') {
            Composition inner;
            bool aaMode_ = s[i] == '[' ? true : aaMode;
            int brace_ = s[i] == '[' ? m_squareBracket : m_parenthesis;
            QChar bracket = s[i];
            errors = checkBrackets(s, sl, &i, j, bracket);

            if (errors == c_OK) {
                // Recurse with string inside the braces, with the information if amino acids
                // mode is enabled and which brace we're in
                errors = qMolCalc::parseString(s.mid(j+1, i-j-1), &inner, aaMode_, brace_);
                j = ++i; // Set indices to after the bracket

                // Check for suffix of brackets
                while (i < sl && ((s[i] <= '9' && s[i] >= '0') || s[i] == '.')) i++;
                double suffix = (i == j ? 1 : s.mid(j, i-j).toDouble(&ok));
                if (! ok ) { //|| suffix == 0) {
                    return QString (QObject::tr("Bad formula.")); // ERROR
                }

                if (suffix == 0) continue; // if suffix==0 don't add the content of the brackets to the composition list

                //inner.applyCoefficient(suffix, aaMode);
                inner.applyCoefficient(suffix, true);

                compTemp.addToList(&(inner));
            } else { return errors; }
        // *******************************************************
        // Check for an element or group symbol
        } else if ((s[i] >= 'A' && s[i] <= 'Z') || (s[i] >= 'a' && s[i] <= 'z')) {
            while (++i < sl && ((s[i] >= 'a' && s[i] <= 'z')
                            || s[i] == '*' || s[i] == '\''
                            || s[i] == c_overriddensymbol[0]));
            QString symbol = s.mid(j, i-j); // extract symbol
            j = i;
            while (i < sl && ((s[i] <= '9' && s[i] >= '0') || s[i] == '.')) i++;
            double suffix = (i == j ? 1 : s.mid(j, i-j).toDouble(&ok));
            if (! ok ) { //|| suffix == 0) {
                return QString(QObject::tr("Bad formula.")); // ERROR
            }

            // *******************************************************
            // Check if 1-letter amino acids mode is active, i.e. only the 1 letter
            // amino acids list will be considered
            if (aaMode) {
                // *******************************************************
                // Check whether the extracted symbol is a 1 letter amino acid
                if (groups->isAA(symbol)) {
                    if (suffix == 0) continue;  // if suffix==0 don't add the current amino acid to the composition list
                    Composition inner;
                    QString groupName = eatWhiteSpace(groups->get1LAminoAcid(symbol));
                    parseString(groupName, &inner);
                    inner.numberAminoAcids += suffix;
                    inner.applyCoefficient(suffix);
                    compTemp.addToList(&(inner));
                } else return QString(QObject::tr("%1 is not an amino acid.").arg(symbol)); // ERROR
            // *******************************************************
            // Check whether the extracted symbol is an element…
            // (note that groups defined with a weight instead of a formula are added
            // to the element table internally and thus treated as if they were an element)
            } else if (elements->contains(symbol)) {
                if (suffix == 0) continue;  // if suffix==0 don't add the current element to the composition list
                compTemp.addElToList(symbol, suffix);
            // *******************************************************
            // … or a defined group
            } else if (groups->contains(symbol)) {
                if (suffix == 0) continue; // if suffix==0 don't add the current group to the composition list
                Composition inner;
                QString groupName = eatWhiteSpace(groups->getFormula(symbol));
                parseString(groupName, &inner);
                inner.applyCoefficient(suffix);
                compTemp.addToList(&(inner));
            } else {
                return QString (QObject::tr("Undefined symbol: ") + symbol); //ERROR
            }
        // a new component after the '+' will be evaluated
        } else if (s[i] == '+' || s[i] == '-') {
            QMap<QString, double>* list = compTemp.getList();
            if (list->isEmpty() && !ok) {
                return QString (QObject::tr("Bad formula.")); // ERROR
            }

            // subtract water if protein mode is on and we're on the square bracket level,
            // because only then a protein is to be finished
            if (proteinMode && brace == m_squareBracket) compTemp.subtractH2O();
            // finish previous component by applying the prefix to it
            compTemp.applyCoefficient(prefix, true);
            comp->addToList(&(compTemp));
            compTemp.clearAA();

            Composition inner;
            // now recurse with the rest of the input string following the +/- (including the sign), then finish
            errors = qMolCalc::parseString(s.mid(i, sl-i), &inner, aaMode, brace);
            comp->addToList(&(inner));

            return errors;
        } else {
            return QString (QObject::tr("Bad formula.")); // ERROR
        }
    }

    // finish the component by applying the prefix to it (multiplying)
    if (prefix == 0) return errors;

    switch (brace) {
        case m_squareBracket:
            if (proteinMode) compTemp.subtractH2O();
        case m_parenthesis:
            compTemp.applyCoefficient(prefix, aaMode);
            break;
        default:
            //compTemp.applyCoefficient(prefix);
            compTemp.applyCoefficient(prefix, true);
    }
    //comp->addToList(compTemp.getList());
    comp->addToList(&(compTemp));
    return errors;
}
