/*
 * qmolelements.cpp
 *
 * This file is part of qmol - A simple molecular weight calculator
 *
 * Copyright (C) 2012 Thomas Mitterfellner
 * Copyright (C) 2009 Tomislav Gountchev
 *
 */

/***************************************************************************
 *                                                                         *
 *   qmol is free software; you can redistribute it and/or modify          *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful, but   *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the Free Software           *
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,                 *
 *   MA 02111-1307 USA                                                     *
 *                                                                         *
 ***************************************************************************/

#include "qmolelements.h"
#include "qmolcalc.h"
#ifdef Q_OS_UNIX
  #include "cmake_defines.h" // this is generated by cmake and defines the installation path
#endif
#include <cmath>
#include <QFile>
#include <QTextStream>
#include <QMessageBox>
#include <QStandardPaths>
#include <QDir>
#include <QDebug>

#ifdef Q_OS_WIN
  #define INSTALLATION_PATH ""
#endif

ElementTable::ElementTable() {
    readTable();
}

ElementTable::~ElementTable() {
}


bool ElementTable::insert(QString name, double weight) {
    bool contained = elTable.contains(name);
    if (contained) override(name);
    elTable.insert(name, weight);
    return contained;
}

void ElementTable::remove(QString name) { elTable.take(name); }

GroupTable::GroupTable(ElementTable* elementTable, qmolPreferences* preferences_) {
    elements = elementTable;
    preferences = preferences_;
    readTables();
}

GroupTable::~GroupTable() {
}

QHash<QString, QString>* GroupTable::getGrpFormulaTable() {
    return &userdefGroups;
}

QHash<QString, double>* GroupTable::getGrpWeightTable() {
    return &grpWeightTable;
}
QList<QString>* GroupTable::getOverriddenTable() {
    return &overriddenSymbols;
}

void ElementTable::readTable() {
    QString name, weight;
    QString installationPath = INSTALLATION_PATH;
    QString elementFileName = "kmolweights";
    QFile elFile(installationPath + elementFileName);

    if (elFile.exists()) {
        elFile.open(QIODevice::ReadOnly);
    } else {
        elFile.setFileName(elementFileName);
        if (elFile.exists()) {
            elFile.open(QIODevice::ReadOnly);
        } else {
            QMessageBox msgBox;
            msgBox.setText(
                "Element weights file " + elementFileName + " not found in:\n" +
                installationPath + "\n" +
                QDir().absolutePath() + "\n" +
                "You will not be able to do any calculations.\n"
                "Consider reinstalling the application or put the file there manually."
            );
            msgBox.exec();
            return;
        }
    }

    elTable.clear();
    QTextStream elFileStr(&elFile);
    while (true) {
        elFileStr >> name >> ws >> weight >> ws;
        if (name == "") break;
        elTable.insert(name, weight.toDouble());
    }
    elFile.close();
}

double ElementTable::getWeight(QString name) const {
    return elTable.value(name);
}

bool ElementTable::contains(QString name) const {
    return elTable.contains(name);
}

void GroupTable::readTables() {
    QList<bool>* activePredefSets = &(preferences->activePredefSets);
    int i = 0;

    elements->readTable();
    // first, read the user's group table
    readTable();

    predefGroups.clear();
    aminoAcids.clear();
    // now read the active predefined groups tables
    foreach (bool predefActive, *activePredefSets) {
       if (predefActive) readTable(predefSetsFiles.at(i));
       i++;
    }

    predefGroups.unite(aminoAcids);
    // now unite both user- and pre-def groups in one pool allGroups
    allGroups.clear();
    allGroups.unite(userdefGroups);
    allGroups.unite(predefGroups);

    overrideUsedElementSymbols();
}

void GroupTable::readTable(const QString tableFile) {
    QString name, defString;
    QHash<QString, QString>* groupsTable;
    QString dataPath = QStandardPaths::standardLocations(QStandardPaths::DataLocation).join("");
    QFile grpFile;
    double groupWeight;
    bool isNumber;
    int i = 0;

    if (!QDir(dataPath).exists()) QDir().mkpath(dataPath);

    // no tableName given means: read the user-defined groups/weights
    if (tableFile == "") {
         grpFile.setFileName(dataPath + "/groups");
         groupsTable = &(userdefGroups);
         groupsTable->clear();
         grpWeightTable.clear();
    // a tableName given means: read a pre-defined groups set
    } else {
         if (tableFile == predefSetsFiles.at(m_AminoAcids1L)) {
             groupsTable = &(aminoAcids1L);
         } else
         if (tableFile == predefSetsFiles.at(m_AminoAcids)) {
             groupsTable = &(aminoAcids);
         } else {
             groupsTable = &(predefGroups);
         }
         grpFile.setFileName(INSTALLATION_PATH + tableFile);
    }

    if (grpFile.exists()) {

    grpFile.open(QIODevice::ReadOnly);
    }

    QTextStream grpFileStream(&grpFile);
    QString line = grpFileStream.readLine();
    // read the file line by line
    while (!line.isNull()) {
        QTextStream lineStream(&line);
        lineStream >> name >> ws;
        int i = name.length()+1;
        defString = line.mid(i);
        //addToGroupTable(name, defString);
        groupWeight = defString.toDouble(&isNumber);

        if (isNumber) { // it is a user-defined weight
            if (elements->insert(name, groupWeight)) addToOverriding(name);
            grpWeightTable.insert(name, groupWeight);
        } else { // it is a group (pre- or user-def)
            groupsTable->insert(name, defString);
        }

        line = grpFileStream.readLine();
    }
    grpFile.close();
}


QString GroupTable::getFormula(QString name) const { return allGroups.value(name); }

QString GroupTable::get1LAminoAcid(QString name) const {
    if (name.length() > 1)
        return aminoAcids.value(name);
    else
        return aminoAcids1L.value(name);
}

bool GroupTable::contains(QString name) const { return (allGroups.contains(name)); }

bool GroupTable::isAA(QString name) const {
    if (name.length() > 1) {
        if (preferences->activePredefSets.at(m_AminoAcids))
            return (aminoAcids.contains(name));
        else
            return false;
    } else {
        return (aminoAcids1L.contains(name));
    }
}

double GroupTable::getWeight(QString name) const {
    return qMolCalc::calculateFormulaWeight(getFormula(name), elements);
}

Composition::Composition() {
    list = new QMap<QString, double>;
    numberAminoAcids = 0;
    numberProteins = 0;
    mustDelete = true;
    suffix=1;
}

Composition::~Composition() {
    if (mustDelete) delete list;
}

void Composition::clear() {
    list->clear();
    numberAminoAcids = numberProteins = 0;
}

void Composition::setList(QMap<QString, double>* comp) {
    QMapIterator<QString, double> listIt(*comp);
    while (listIt.hasNext()) {
        listIt.next();
        QString key = listIt.key();
        double value = listIt.value();
        list->insert(key, value);
    }
}

int Composition::getNumberAminoAcids() { return numberAminoAcids; }
void Composition::clearAA() { numberAminoAcids = 0; }
int Composition::getNumberProteins() { return numberProteins; }

void Composition::addAminoAcid(int number) { numberAminoAcids += number; };
void Composition::addProtein(int number) { numberProteins += number; };

// add a list of elements to the current element list
//void Composition::addToList(QMap<QString, double>* comp) {
void Composition::addToList(Composition* comp_) {
    QMap<QString, double>* comp = comp_->getList();
    int numberAminoAcids_ = comp_->getNumberAminoAcids();
    int numberProteins_ = comp_->getNumberProteins();
    numberAminoAcids += numberAminoAcids_;
    //numberProteins += numberProteins_;
    QMapIterator<QString, double> listIt(*comp);
    while (listIt.hasNext()) {
        listIt.next();
        QString key = listIt.key();
        double value = listIt.value();
        if (list->contains(key)) {
            (*list)[key] += value;
        } else {
            list->insert(key, value);
        }
    }
}

void Composition::subtractH2O() {
    if (numberAminoAcids != 0) {
        int numberH2O = abs(numberAminoAcids) - 1;
        int coefficient = list->take("H");
        coefficient -= 2 * numberH2O;
        list->insert("H", coefficient);
        coefficient = list->take("O");
        coefficient -= numberH2O;
        list->insert("O", coefficient);
    }
}


QMap<QString, double>* Composition::getList() {
    return list;
}

/* add element symbol to the current composition list
 * (only add to the coefficient if already present) */
void Composition::addElToList(QString name, double coef) {
    if (list->contains(name)) {
        (*list)[name]+= coef;
    } else {
        list->insert(name, coef);
    }
}

// Multiply each member of the current composition list with the passed coefficient
void Composition::applyCoefficient(double coeff, bool multAA) {
    QMutableMapIterator<QString, double> listIt(*list);
    suffix = coeff;

    while (listIt.hasNext()) {
        listIt.next();
        listIt.value() *= suffix;
    }
    // multiply the amino acids only if explicitly told to do so
    if (multAA) numberAminoAcids *= coeff;
}

void GroupTable::addToGroupFormulaTable(QString name, QString compString) {
    userdefGroups.insert(name, compString);
}

/* Checks all group symbols whether they clash with an element symbol
   If so, override the original element symbol */
void GroupTable::overrideUsedElementSymbols() {
    overriddenSymbols.clear();
    foreach (QString groupSymbol, allGroups.keys()) {
        if (elements->contains(groupSymbol)) {
            elements->override(groupSymbol);
            overriddenSymbols.append(groupSymbol);
        }
    }
}

void GroupTable::addToOverriding(QString name) { overriddenSymbols.append(name); }

bool GroupTable::isOverriding(QString name) { return overriddenSymbols.contains(name); }

/* Overrides an element symbol. Effectively, this is renaming the
   original 'Symbol' to 'Symbol!', by which it will be accessible then */
void ElementTable::override(const QString name) {
    double oldWeight = elTable.take(name);
    elTable.insert(name + c_overriddensymbol, oldWeight);
}

/* Reverses overriding an element symbol. Effectively, this is renaming the
   'Symbol!' to the original 'Symbol', plus overwriting the formerly defined symbol */
void ElementTable::undoOverride(const QString name) {
    double oldWeight = elTable.take(name + c_overriddensymbol);
    elTable.insert(name, oldWeight);
}
